@@ -69,7 +69,7 @@ VAR libstackv library_stack;
 #endif
 
 //int IsCmd(char *n, int tok);
-char mytolower(char c);
+static char mytolower(char c);
 
 /*2
 * return TRUE if the libray libname is already loaded
@@ -205,7 +205,8 @@ char* iiGetLibProcBuffer(procinfo *pi, int part )
     return NULL;
   }
 
-  fseek(fp, pi->data.s.proc_start, SEEK_SET);
+  int res=fseek(fp, pi->data.s.proc_start, SEEK_SET);
+  if (res==1) return NULL;
   if(part==0)
   { // load help string
     int i, offset=0;
@@ -219,10 +220,13 @@ char* iiGetLibProcBuffer(procinfo *pi, int part )
     //Print("Help=%ld-%ld=%d\n", pi->data.s.body_start,
     //    pi->data.s.proc_start, procbuflen);
     s = (char *)omAlloc(procbuflen+head+3);
-    myfread(s, head, 1, fp);
+    res=myfread(s, head, 1, fp);
+    if (res<=0) /* error*/ { omFree(s); return NULL; }
     s[head] = '\n';
-    fseek(fp, pi->data.s.help_start, SEEK_SET);
-    myfread(s+head+1, procbuflen, 1, fp);
+    int res=fseek(fp, pi->data.s.help_start, SEEK_SET);
+    if (res==-1) /*error*/ { omFree(s); return NULL; }
+    res=myfread(s+head+1, procbuflen, 1, fp);
+    if (res<=0) /* error*/ { omFree(s); return NULL; }
     fclose(fp);
     s[procbuflen+head+1] = '\n';
     s[procbuflen+head+2] = '\0';
@@ -280,7 +284,11 @@ char* iiGetLibProcBuffer(procinfo *pi, int part )
       return NULL; // example part does not exist
     // load example
     fseek(fp, pi->data.s.example_start, SEEK_SET);
-    /*char *dummy=*/ (void) fgets(buf, sizeof(buf), fp); // skip line with "example"
+    char* res=fgets(buf, sizeof(buf), fp); // skip line with "example"
+    if (res==NULL) /*error or eof*/
+    {
+      return omStrDup("");
+    }
     procbuflen = pi->data.s.proc_end - pi->data.s.example_start - strlen(buf);
     //Print("Example=%ld-%ld=%d\n", pi->data.s.proc_end,
     //  pi->data.s.example_start, procbuflen);
@@ -824,7 +832,8 @@ BOOLEAN iiTryLoadLib(leftv v, const char *id)
 {
   BOOLEAN LoadResult = TRUE;
   char libnamebuf[1024];
-  char *libname = (char *)omAlloc(strlen(id)+5);
+  size_t len=strlen(id)+5;
+  char *libname = (char *)omAlloc(len);
   const char *suffix[] = { "", ".lib", ".so", ".sl", NULL };
   int i = 0;
   // FILE *fp;
@@ -833,7 +842,7 @@ BOOLEAN iiTryLoadLib(leftv v, const char *id)
   lib_types LT;
   for(i=0; suffix[i] != NULL; i++)
   {
-    sprintf(libname, "%s%s", id, suffix[i]);
+    snprintf(libname,len, "%s%s", id, suffix[i]);
     *libname = mytolower(*libname);
     if((LT = type_of_LIB(libname, libnamebuf)) > LT_NOTFOUND)
     {
@@ -1194,7 +1203,7 @@ BOOLEAN load_modules_aux(const char *newlib, char *fullname, BOOLEAN autoexport)
   char *FullName=(char*)omAlloc0(l);
 
   if( *fullname != '/' &&  *fullname != '.' )
-    sprintf(FullName, "./%s", newlib);
+    snprintf(FullName,l, "./%s", newlib);
   else strncpy(FullName, fullname,l);
 
 
@@ -1391,7 +1400,7 @@ void* binary_module_function(const char* newlib, const char* funcname)
   if (!bin_dir)  { return NULL; }
 
   char path_name[MAXPATHLEN];
-  sprintf(path_name, "%s%s%s.%s", bin_dir, DIR_SEPP, newlib, MODULE_SUFFIX_STRING);
+  snprintf(path_name,MAXPATHLEN, "%s%s%s.%s", bin_dir, DIR_SEPP, newlib, MODULE_SUFFIX_STRING);
 
   void* openlib = dynl_open(path_name);
   if(!openlib)
@@ -1407,13 +1416,13 @@ void* binary_module_function(const char* newlib, const char* funcname)
 #endif
 
 /*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
-char mytoupper(char c)
+static char mytoupper(char c)
 {
   if(c>=97 && c<=(97+26)) c-=32;
   return(c);
 }
 
-char mytolower(char c)
+static char mytolower(char c)
 {
   if(c>=65 && c<=(65+26)) c+=32;
   return(c);
@@ -1463,7 +1472,8 @@ void piShowProcList()
       if(strcmp(proc->procname, IDID(h))!=0)
       {
         name = (char *)omAlloc(strlen(IDID(h))+strlen(proc->procname)+4);
-        sprintf(name, "%s -> %s", IDID(h), proc->procname);
+        snprintf(name, strlen(IDID(h))+strlen(proc->procname)+4,
+	     "%s -> %s", IDID(h), proc->procname);
         Print( "%d %-15s  %20s ", proc->is_static ? 1 : 0, proc->libname, name);
         omFree((ADDRESS)name);
       }
@@ -1489,8 +1499,8 @@ void piShowProcList()
 //  idhdl pn = ggetid(procname);
 //  procinfo *pi = IDPROC(pn);
 //
-//  sprintf(buf, "%s %3d\0", procname, lineno);
-//  //sprintf(buf, "%s::%s %3d\0", pi->libname, pi->procname,
+//  snprintf(buf,256, "%s %3d\0", procname, lineno);
+//  //snprintf(buf,256, "%s::%s %3d\0", pi->libname, pi->procname,
 //  //  lineno + pi->data.s.body_lineno);
 //  return(buf);
 //}
