@@ -49,7 +49,7 @@
 #include <netdb.h>
 #include <netinet/in.h> /* for htons etc.*/
 
-#define SSI_VERSION 13
+#define SSI_VERSION 15
 // 5->6: changed newstruct representation
 // 6->7: attributes
 // 7->8: qring
@@ -58,18 +58,21 @@
 // 10->11: extended ring descr. for named coeffs (not in used until 4.1)
 // 11->12: add rank to ideal/module, add smatrix
 // 12->13: NC rings
+// 13->14: ring references
+// 14->15: bigintvec, prune_map, mres_map
 
 VAR link_list ssiToBeClosed=NULL;
 VAR volatile BOOLEAN ssiToBeClosed_inactive=TRUE;
 
 // forward declarations:
-void ssiWritePoly_R(const ssiInfo *d, int typ, poly p, const ring r);
-void ssiWriteIdeal_R(const ssiInfo *d, int typ,const ideal I, const ring r);
-poly ssiReadPoly_R(const ssiInfo *D, const ring r);
-ideal ssiReadIdeal_R(const ssiInfo *d,const ring r);
+static void ssiWriteIdeal(const ssiInfo *d, int typ,const ideal I);
+static void ssiWritePoly_R(const ssiInfo *d, int typ, poly p, const ring r);
+static void ssiWriteIdeal_R(const ssiInfo *d, int typ,const ideal I, const ring r);
+static poly ssiReadPoly_R(const ssiInfo *D, const ring r);
+static ideal ssiReadIdeal_R(const ssiInfo *d,const ring r);
 
 // the helper functions:
-BOOLEAN ssiSetCurrRing(const ring r) /* returned: not accepted */
+static BOOLEAN ssiSetCurrRing(const ring r) /* returned: not accepted */
 {
   //  if (currRing!=NULL)
   //  Print("need to change the ring, currRing:%s, switch to: ssiRing%d\n",IDID(currRingHdl),nr);
@@ -88,7 +91,7 @@ BOOLEAN ssiSetCurrRing(const ring r) /* returned: not accepted */
     idhdl h=NULL;
     loop
     {
-      sprintf(name,"ssiRing%d",nr); nr++;
+      snprintf(name,20,"ssiRing%d",nr); nr++;
       h=IDROOT->get(name, 0);
       if (h==NULL)
       {
@@ -114,7 +117,7 @@ BOOLEAN ssiSetCurrRing(const ring r) /* returned: not accepted */
     return TRUE;
   }
 }
-void ssiCheckCurrRing(const ring r)
+static void ssiCheckCurrRing(const ring r)
 {
   if ((r!=currRing)
   ||(currRingHdl==NULL)
@@ -125,7 +128,7 @@ void ssiCheckCurrRing(const ring r)
     idhdl h=NULL;
     loop
     {
-      sprintf(name,"ssiRing%d",nr); nr++;
+      snprintf(name,20,"ssiRing%d",nr); nr++;
       h=IDROOT->get(name, 0);
       if (h==NULL)
       {
@@ -145,24 +148,24 @@ void ssiCheckCurrRing(const ring r)
   assume((currRing==r) || rEqual(r,currRing));
 }
 // the implementation of the functions:
-void ssiWriteInt(const ssiInfo *d,const int i)
+static void ssiWriteInt(const ssiInfo *d,const int i)
 {
   fprintf(d->f_write,"%d ",i);
   //if (d->f_debug!=NULL) fprintf(d->f_debug,"int: %d ",i);
 }
 
-void ssiWriteString(const ssiInfo *d,const char *s)
+static void ssiWriteString(const ssiInfo *d,const char *s)
 {
   fprintf(d->f_write,"%d %s ",(int)strlen(s),s);
   //if (d->f_debug!=NULL) fprintf(d->f_debug,"stringi: %d \"%s\" ",strlen(s),s);
 }
 
-void ssiWriteBigInt(const ssiInfo *d, const number n)
+static void ssiWriteBigInt(const ssiInfo *d, const number n)
 {
  n_WriteFd(n,d,coeffs_BIGINT);
 }
 
-void ssiWriteNumber_CF(const ssiInfo *d, const number n, const coeffs cf)
+static void ssiWriteNumber_CF(const ssiInfo *d, const number n, const coeffs cf)
 {
   // syntax is as follows:
   // case 1 Z/p:   3 <int>
@@ -190,19 +193,39 @@ void ssiWriteNumber_CF(const ssiInfo *d, const number n, const coeffs cf)
   else WerrorS("coeff field not implemented");
 }
 
-void ssiWriteNumber(const ssiInfo *d, const number n)
+static void ssiWriteNumber(const ssiInfo *d, const number n)
 {
   ssiWriteNumber_CF(d,n,d->r->cf);
 }
 
-void ssiWriteRing_R(ssiInfo *d,const ring r)
+static void ssiWriteRing_R(ssiInfo *d,const ring r)
 {
   /* 5 <ch> <N> <l1> <v1> ...<lN> <vN> <number of orderings> <ord1> <block0_1> <block1_1> .... <extRing> <Q-ideal> */
   /* ch=-1: transext, coeff ring follows */
   /* ch=-2: algext, coeff ring and minpoly follows */
   /* ch=-3: cf name follows */
+  /* ch=-4: NULL*/
+  /* ch=-5: reference <int> */
+  /* ch=-6: new reference <int> <ring> */
   if (r!=NULL)
   {
+    for(int i=0;i<SI_RING_CACHE;i++)
+    {
+      if (d->rings[i]==r)
+      {
+        fprintf(d->f_write,"-5 %d ",i);
+        return;
+      }
+    }
+    for(int i=0;i<SI_RING_CACHE;i++)
+    {
+      if (d->rings[i]==NULL)
+      {
+        d->rings[i]=rIncRefCnt(r);
+        fprintf(d->f_write,"-6 %d ",i);
+        break;
+      }
+    }
     if (rField_is_Q(r) || rField_is_Zp(r))
       fprintf(d->f_write,"%d %d ",n_GetChar(r->cf),r->N);
     else if (rFieldType(r)==n_transExt)
@@ -239,14 +262,22 @@ void ssiWriteRing_R(ssiInfo *d,const ring r)
         case ringorder_Ws:
         case ringorder_aa:
         {
-          int ii;
-          for(ii=r->block0[i];ii<=r->block1[i];ii++)
-            fprintf(d->f_write,"%d ",r->wvhdl[i][ii-r->block0[i]]);
+          int s=r->block1[i]-r->block0[i]+1; // #vars
+          for(int ii=0;ii<s;ii++)
+            fprintf(d->f_write,"%d ",r->wvhdl[i][ii]);
+        }
+        break;
+        case ringorder_M:
+        {
+          int s=r->block1[i]-r->block0[i]+1; // #vars
+          for(int ii=0;ii<s*s;ii++)
+          {
+            fprintf(d->f_write,"%d ",r->wvhdl[i][ii]);
+          }
         }
         break;
 
         case ringorder_a64:
-        case ringorder_M:
         case ringorder_L:
         case ringorder_IS:
           Werror("ring oder not implemented for ssi:%d",r->order[i]);
@@ -291,20 +322,21 @@ void ssiWriteRing_R(ssiInfo *d,const ring r)
     if (rIsPluralRing(r))
     {
       fputs("23 2 ",d->f_write);
-      void ssiWriteIdeal(const ssiInfo *d, int typ,const ideal I);
       ssiWriteIdeal(d,MATRIX_CMD,(ideal)r->GetNC()->C);
       ssiWriteIdeal(d,MATRIX_CMD,(ideal)r->GetNC()->D);
     }
   }
 }
 
-void ssiWriteRing(ssiInfo *d,const ring r)
+static void ssiWriteRing(ssiInfo *d,const ring r)
 {
   /* 5 <ch> <N> <l1> <v1> ...<lN> <vN> <number of orderings> <ord1> <block0_1> <block1_1> .... <extRing> <Q-ideal> */
   /* ch=-1: transext, coeff ring follows */
   /* ch=-2: algext, coeff ring and minpoly follows */
   /* ch=-3: cf name follows */
   /* ch=-4: NULL */
+  /* ch=-5: reference <int> */
+  /* ch=-6: new reference <int> <ring> */
   if ((r==NULL)||(r->cf==NULL))
   {
     fputs("-4 ",d->f_write);
@@ -321,7 +353,7 @@ void ssiWriteRing(ssiInfo *d,const ring r)
   }
   ssiWriteRing_R(d,r);
 }
-void ssiWritePoly_R(const ssiInfo *d, int /*typ*/, poly p, const ring r)
+static void ssiWritePoly_R(const ssiInfo *d, int /*typ*/, poly p, const ring r)
 {
   fprintf(d->f_write,"%d ",pLength(p));//number of terms
 
@@ -339,12 +371,12 @@ void ssiWritePoly_R(const ssiInfo *d, int /*typ*/, poly p, const ring r)
   }
 }
 
-void ssiWritePoly(const ssiInfo *d, int typ, poly p)
+static void ssiWritePoly(const ssiInfo *d, int typ, poly p)
 {
   ssiWritePoly_R(d,typ,p,d->r);
 }
 
-void ssiWriteIdeal_R(const ssiInfo *d, int typ,const ideal I, const ring R)
+static void ssiWriteIdeal_R(const ssiInfo *d, int typ,const ideal I, const ring R)
 {
    // syntax: 7 # of elements <poly 1> <poly2>.....(ideal,module,smatrix)
    // syntax: 8 <rows> <cols> <poly 1> <poly2>.....(matrix)
@@ -374,12 +406,12 @@ void ssiWriteIdeal_R(const ssiInfo *d, int typ,const ideal I, const ring R)
      ssiWritePoly_R(d,tt,I->m[i],R);
    }
 }
-void ssiWriteIdeal(const ssiInfo *d, int typ,const ideal I)
+static void ssiWriteIdeal(const ssiInfo *d, int typ,const ideal I)
 {
   ssiWriteIdeal_R(d,typ,I,d->r);
 }
 
-void ssiWriteCommand(si_link l, command D)
+static void ssiWriteCommand(si_link l, command D)
 {
   ssiInfo *d=(ssiInfo*)l->data;
   // syntax: <num ops> <operation> <op1> <op2> ....
@@ -392,7 +424,7 @@ void ssiWriteCommand(si_link l, command D)
   }
 }
 
-void ssiWriteProc(const ssiInfo *d,procinfov p)
+static void ssiWriteProc(const ssiInfo *d,procinfov p)
 {
   if (p->data.s.body==NULL)
     iiGetLibProcBuffer(p);
@@ -402,7 +434,7 @@ void ssiWriteProc(const ssiInfo *d,procinfov p)
     ssiWriteString(d,"");
 }
 
-void ssiWriteList(si_link l,lists dd)
+static void ssiWriteList(si_link l,lists dd)
 {
   ssiInfo *d=(ssiInfo*)l->data;
   int Ll=dd->nr;
@@ -413,7 +445,7 @@ void ssiWriteList(si_link l,lists dd)
     ssiWrite(l,&(dd->m[i]));
   }
 }
-void ssiWriteIntvec(const ssiInfo *d,intvec * v)
+static void ssiWriteIntvec(const ssiInfo *d,intvec * v)
 {
   fprintf(d->f_write,"%d ",v->length());
   int i;
@@ -422,7 +454,7 @@ void ssiWriteIntvec(const ssiInfo *d,intvec * v)
     fprintf(d->f_write,"%d ",(*v)[i]);
   }
 }
-void ssiWriteIntmat(const ssiInfo *d,intvec * v)
+static void ssiWriteIntmat(const ssiInfo *d,intvec * v)
 {
   fprintf(d->f_write,"%d %d ",v->rows(),v->cols());
   int i;
@@ -432,7 +464,7 @@ void ssiWriteIntmat(const ssiInfo *d,intvec * v)
   }
 }
 
-void ssiWriteBigintmat(const ssiInfo *d,bigintmat * v)
+static void ssiWriteBigintmat(const ssiInfo *d,bigintmat * v)
 {
   fprintf(d->f_write,"%d %d ",v->rows(),v->cols());
   int i;
@@ -442,7 +474,17 @@ void ssiWriteBigintmat(const ssiInfo *d,bigintmat * v)
   }
 }
 
-char *ssiReadString(const ssiInfo *d)
+static void ssiWriteBigintvec(const ssiInfo *d,bigintmat * v)
+{
+  fprintf(d->f_write,"%d ",v->cols());
+  int i;
+  for(i=0;i<v->length();i++)
+  {
+    ssiWriteBigInt(d,(*v)[i]);
+  }
+}
+
+static char *ssiReadString(const ssiInfo *d)
 {
   char *buf;
   int l;
@@ -455,14 +497,14 @@ char *ssiReadString(const ssiInfo *d)
   return buf;
 }
 
-int ssiReadInt(s_buff fich)
+static int ssiReadInt(s_buff fich)
 {
   return s_readint(fich);
 }
 
-number ssiReadNumber_CF(const ssiInfo *d, const coeffs cf)
+static number ssiReadNumber_CF(const ssiInfo *d, const coeffs cf)
 {
-  if (cf->cfReadFd!=NULL)
+  if (cf->cfReadFd!=ndReadFd)
   {
      return n_ReadFd(d,cf);
   }
@@ -484,7 +526,7 @@ number ssiReadNumber_CF(const ssiInfo *d, const coeffs cf)
   return NULL;
 }
 
-number ssiReadBigInt(const ssiInfo *d)
+static number ssiReadBigInt(const ssiInfo *d)
 {
   number n=ssiReadNumber_CF(d,coeffs_BIGINT);
   if ((SR_HDL(n) & SR_INT)==0)
@@ -494,16 +536,29 @@ number ssiReadBigInt(const ssiInfo *d)
   return n;
 }
 
-number ssiReadNumber(ssiInfo *d)
+static number ssiReadNumber(ssiInfo *d)
 {
   return ssiReadNumber_CF(d,d->r->cf);
 }
 
-ring ssiReadRing(const ssiInfo *d)
+static ring ssiReadRing(ssiInfo *d)
 {
 /* syntax is <ch> <N> <l1> <v1> ...<lN> <vN> <number of orderings> <ord1> <block0_1> <block1_1> .... <Q-ideal> */
   int ch;
+  int new_ref=-1;
   ch=s_readint(d->f_read);
+  if (ch==-6)
+  {
+    new_ref=s_readint(d->f_read);
+    ch=s_readint(d->f_read);
+  }
+  if (ch==-5)
+  {
+    int index=s_readint(d->f_read);
+    ring r=d->rings[index];
+    rIncRefCnt(r);
+    return r;
+  }
   if (ch==-4)
     return NULL;
   int N=s_readint(d->f_read);
@@ -516,7 +571,7 @@ ring ssiReadRing(const ssiInfo *d)
     if (cf==NULL)
     {
       Werror("cannot find cf:%s",cf_name);
-      omFree(cf_name);
+      omFreeBinAddr(cf_name);
       return NULL;
     }
   }
@@ -549,18 +604,26 @@ ring ssiReadRing(const ssiInfo *d)
       case ringorder_Ws:
       case ringorder_aa:
       {
-        wvhdl[i]=(int*)omAlloc((block1[i]-block0[i]+1)*sizeof(int));
-        int ii;
-        for(ii=block0[i];ii<=block1[i];ii++)
-          wvhdl[i][ii-block0[i]]=s_readint(d->f_read);
+        int s=block1[i]-block0[i]+1; // #vars
+        wvhdl[i]=(int*)omAlloc(s*sizeof(int));
+        for(int ii=0;ii<s;ii++)
+          wvhdl[i][ii]=s_readint(d->f_read);
       }
       break;
-
-      case ringorder_a64:
       case ringorder_M:
+      {
+        int s=block1[i]-block0[i]+1; // #vars
+        wvhdl[i]=(int*)omAlloc(s*s*sizeof(int));
+        for(int ii=0;ii<s*s;ii++)
+        {
+          wvhdl[i][ii]=s_readint(d->f_read);
+        }
+      }
+      break;
+      case ringorder_a64:
       case ringorder_L:
       case ringorder_IS:
-        Werror("ring oder not implemented for ssi:%d",ord[i]);
+        Werror("ring order not implemented for ssi:%d",ord[i]);
         break;
 
       default: break;
@@ -624,7 +687,7 @@ ring ssiReadRing(const ssiInfo *d)
     idhdl h=NULL;
     loop
     {
-      sprintf(name,"ssiRing%d",nr); nr++;
+      snprintf(name,20,"ssiRing%d",nr); nr++;
       h=IDROOT->get(name, 0);
       if (h==NULL)
       {
@@ -634,16 +697,21 @@ ring ssiReadRing(const ssiInfo *d)
       && (r!=IDRING(h))
       && (rEqual(r,IDRING(h),1)))
       {
-	rDelete(r);
+        rDelete(r);
         r=rIncRefCnt(IDRING(h));
         break;
       }
     }
+    if (new_ref!=-1)
+    {
+      d->rings[new_ref]=r;
+      rIncRefCnt(r);
+    }
     return r;
   }
 }
 
-poly ssiReadPoly_R(const ssiInfo *d, const ring r)
+static poly ssiReadPoly_R(const ssiInfo *d, const ring r)
 {
 // < # of terms> < term1> < .....
   int n,i,l;
@@ -674,12 +742,12 @@ poly ssiReadPoly_R(const ssiInfo *d, const ring r)
  return ret;
 }
 
-poly ssiReadPoly(ssiInfo *d)
+static poly ssiReadPoly(ssiInfo *d)
 {
   return ssiReadPoly_R(d,d->r);
 }
 
-ideal ssiReadIdeal_R(const ssiInfo *d,const ring r)
+static ideal ssiReadIdeal_R(const ssiInfo *d,const ring r)
 {
 // < # of terms> < term1> < .....
   int n,i;
@@ -693,12 +761,12 @@ ideal ssiReadIdeal_R(const ssiInfo *d,const ring r)
   return I;
 }
 
-ideal ssiReadIdeal(ssiInfo *d)
+static ideal ssiReadIdeal(ssiInfo *d)
 {
   return ssiReadIdeal_R(d,d->r);
 }
 
-matrix ssiReadMatrix(ssiInfo *d)
+static matrix ssiReadMatrix(ssiInfo *d)
 {
   int n,m;
   m=s_readint(d->f_read);
@@ -714,7 +782,7 @@ matrix ssiReadMatrix(ssiInfo *d)
   return M;
 }
 
-command ssiReadCommand(si_link l)
+static command ssiReadCommand(si_link l)
 {
   ssiInfo *d=(ssiInfo*)l->data;
   // syntax: <num ops> <operation> <op1> <op2> ....
@@ -760,7 +828,7 @@ command ssiReadCommand(si_link l)
   return D;
 }
 
-procinfov ssiReadProc(const ssiInfo *d)
+static procinfov ssiReadProc(const ssiInfo *d)
 {
   char *s=ssiReadString(d);
   procinfov p=(procinfov)omAlloc0Bin(procinfo_bin);
@@ -770,7 +838,7 @@ procinfov ssiReadProc(const ssiInfo *d)
   p->data.s.body=s;
   return p;
 }
-lists ssiReadList(si_link l)
+static lists ssiReadList(si_link l)
 {
   ssiInfo *d=(ssiInfo*)l->data;
   int nr;
@@ -788,7 +856,7 @@ lists ssiReadList(si_link l)
   }
   return L;
 }
-intvec* ssiReadIntvec(const ssiInfo *d)
+static intvec* ssiReadIntvec(const ssiInfo *d)
 {
   int nr;
   nr=s_readint(d->f_read);
@@ -799,7 +867,7 @@ intvec* ssiReadIntvec(const ssiInfo *d)
   }
   return v;
 }
-intvec* ssiReadIntmat(const ssiInfo *d)
+static intvec* ssiReadIntmat(const ssiInfo *d)
 {
   int r,c;
   r=s_readint(d->f_read);
@@ -811,7 +879,7 @@ intvec* ssiReadIntmat(const ssiInfo *d)
   }
   return v;
 }
-bigintmat* ssiReadBigintmat(const ssiInfo *d)
+static bigintmat* ssiReadBigintmat(const ssiInfo *d)
 {
   int r,c;
   r=s_readint(d->f_read);
@@ -823,8 +891,19 @@ bigintmat* ssiReadBigintmat(const ssiInfo *d)
   }
   return v;
 }
+static bigintmat* ssiReadBigintvec(const ssiInfo *d)
+{
+  int c;
+  c=s_readint(d->f_read);
+  bigintmat *v=new bigintmat(1,c,coeffs_BIGINT);
+  for(int i=0;i<c;i++)
+  {
+    (*v)[i]=ssiReadBigInt(d);
+  }
+  return v;
+}
 
-void ssiReadBlackbox(leftv res, si_link l)
+static void ssiReadBlackbox(leftv res, si_link l)
 {
   ssiInfo *d=(ssiInfo*)l->data;
   int throwaway=s_readint(d->f_read);
@@ -852,7 +931,7 @@ void ssiReadBlackbox(leftv res, si_link l)
   omFree(name);
 }
 
-void ssiReadAttrib(leftv res, si_link l)
+static void ssiReadAttrib(leftv res, si_link l)
 {
   ssiInfo *d=(ssiInfo*)l->data;
   BITSET fl=(BITSET)s_readint(d->f_read);
@@ -872,7 +951,7 @@ void ssiReadAttrib(leftv res, si_link l)
   }
   res->flag=fl;
 }
-void ssiReadRingProperties(si_link l)
+static void ssiReadRingProperties(si_link l)
 {
   ssiInfo *d=(ssiInfo*)l->data;
   int what=s_readint(d->f_read);
@@ -937,23 +1016,28 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
     SI_LINK_SET_OPEN_P(l, flag);
     if(l->data!=NULL) omFreeSize(l->data,sizeof(ssiInfo));
     l->data=d;
-    omFree(l->mode);
+    omFreeBinAddr(l->mode);
     l->mode = omStrDup(mode);
 
     if (l->name[0] == '\0')
     {
       if (strcmp(mode,"fork")==0)
       {
+        int pc[2];
+        int cp[2];
+        int err1=pipe(pc);
+        int err2=pipe(cp);
+        if (err1 || err2)
+        {
+          Werror("pipe failed with %d\n",errno);
+          return TRUE;
+        }
         link_list n=(link_list)omAlloc(sizeof(link_struct));
         n->u=u;
         n->l=l;
         n->next=(void *)ssiToBeClosed;
         ssiToBeClosed=n;
 
-        int pc[2];
-        int cp[2];
-        pipe(pc);
-        pipe(cp);
         pid_t pid = fork();
         if (pid == -1 && errno == EAGAIN)   // RLIMIT_NPROC too low?
         {
@@ -1001,7 +1085,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
           //d->r=currRing;
           //if (d->r!=NULL) d->r->ref++;
           l->data=d;
-          omFree(l->mode);
+          omFreeBinAddr(l->mode);
           l->mode = omStrDup(mode);
           singular_in_batchmode=TRUE;
           SI_LINK_SET_RW_OPEN_P(l);
@@ -1060,7 +1144,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
         {
           WerrorS("ERROR opening socket");
           l->data=NULL;
-          SI_LINK_CLOSE_P(l);
+          l->flags=0;
           omFree(d);
           return TRUE;
         }
@@ -1076,7 +1160,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
           {
             WerrorS("ERROR on binding (no free port available?)");
             l->data=NULL;
-            SI_LINK_CLOSE_P(l);
+            l->flags=0;
             omFree(d);
             return TRUE;
           }
@@ -1089,7 +1173,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
         {
           WerrorS("ERROR on accept");
           l->data=NULL;
-          SI_LINK_CLOSE_P(l);
+          l->flags=0;
           omFree(d);
           return TRUE;
         }
@@ -1106,13 +1190,13 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
       {
         Werror("invalid mode >>%s<< for ssi",mode);
         l->data=NULL;
-        SI_LINK_CLOSE_P(l);
+        l->flags=0;
         omFree(d);
         return TRUE;
       }
     }
     // =========================================================================
-    else /*l->name=NULL*/
+    else /*now l->name!=NULL*/
     {
       // tcp mode
       if(strcmp(mode,"tcp")==0)
@@ -1124,7 +1208,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
         {
           WerrorS("ERROR opening socket");
           l->data=NULL;
-          SI_LINK_CLOSE_P(l);
+          l->flags=0;
           omFree(d);
           return TRUE;
         }
@@ -1140,7 +1224,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
           {
             WerrorS("ERROR on binding (no free port available?)");
             l->data=NULL;
-            SI_LINK_CLOSE_P(l);
+            l->flags=0;
             return TRUE;
           }
         }
@@ -1154,7 +1238,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
         {
           WerrorS("ERROR: no host specified");
           l->data=NULL;
-          SI_LINK_CLOSE_P(l);
+          l->flags=0;
           omFree(d);
           omFree(path);
           omFree(cli_host);
@@ -1168,16 +1252,27 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
         }
         char* ssh_command = (char*)omAlloc(256);
         char* ser_host = (char*)omAlloc(64);
-        gethostname(ser_host,64);
-        if (strcmp(cli_host,"localhost")==0) /*avoid "ssh localhost" as key may change*/
-          sprintf(ssh_command,"%s -q --batch --link=ssi --MPhost=%s --MPport=%d &",path,ser_host,portno);
+        if(strcmp(cli_host,"localhost")==0)
+          strcpy(ser_host,"localhost");
         else
-          sprintf(ssh_command,"ssh %s %s -q --batch --link=ssi --MPhost=%s --MPport=%d &",cli_host,path,ser_host,portno);
+          gethostname(ser_host,64);
+        if (strcmp(cli_host,"localhost")==0) /*avoid "ssh localhost" as key may change*/
+          snprintf(ssh_command,256,"%s -q --batch --link=ssi --MPhost=%s --MPport=%d &",path,ser_host,portno);
+        else
+          snprintf(ssh_command,256,"ssh %s %s -q --batch --link=ssi --MPhost=%s --MPport=%d &",cli_host,path,ser_host,portno);
         //Print("client on %s started:%s\n",cli_host,path);
         omFree(path);
         omFree(cli_host);
         if (TEST_OPT_PROT) { Print("running >>%s<<\n",ssh_command); }
-        system(ssh_command);
+        int re=system(ssh_command);
+        if (re<0)
+        {
+          Werror("ERROR running `%s` (%d)",ssh_command,re);
+          l->data=NULL;
+          l->flags=0;
+          omFree(d);
+          return TRUE;
+        }
         omFree(ssh_command);
         omFree(ser_host);
         clilen = sizeof(cli_addr);
@@ -1186,7 +1281,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
         {
           WerrorS("ERROR on accept");
           l->data=NULL;
-          SI_LINK_CLOSE_P(l);
+          l->flags=0;
           omFree(d);
           return TRUE;
         }
@@ -1221,14 +1316,14 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
           if (sockfd < 0)
           {
             WerrorS("ERROR opening socket");
-            SI_LINK_CLOSE_P(l);
+            l->flags=0;
             return TRUE;
           }
           server = gethostbyname(host);
           if (server == NULL)
           {
             WerrorS("ERROR, no such host");
-            SI_LINK_CLOSE_P(l);
+            l->flags=0;
             return TRUE;
           }
           memset((char *) &serv_addr, 0, sizeof(serv_addr));
@@ -1240,7 +1335,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
           if (si_connect(sockfd,(sockaddr*)&serv_addr,sizeof(serv_addr)) < 0)
           {
             Werror("ERROR connecting(errno=%d)",errno);
-            SI_LINK_CLOSE_P(l);
+            l->flags=0;
             return TRUE;
           }
           //PrintS("connected\n");mflush();
@@ -1254,7 +1349,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
         else
         {
           l->data=NULL;
-          SI_LINK_CLOSE_P(l);
+          l->flags=0;
           omFree(d);
           return TRUE;
         }
@@ -1297,7 +1392,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
         {
           omFree(d);
           l->data=NULL;
-          SI_LINK_CLOSE_P(l);
+          l->flags=0;
           return TRUE;
         }
       }
@@ -1308,7 +1403,7 @@ BOOLEAN ssiOpen(si_link l, short flag, leftv u)
 }
 
 //**************************************************************************/
-BOOLEAN ssiPrepClose(si_link l)
+static BOOLEAN ssiPrepClose(si_link l)
 {
   if (l!=NULL)
   {
@@ -1344,6 +1439,11 @@ BOOLEAN ssiClose(si_link l)
       }
       // clean ring
       if (d->r!=NULL) rKill(d->r);
+      for(int i=0;i<SI_RING_CACHE;i++)
+      {
+        if (d->rings[i]!=NULL)  rKill(d->rings[i]);
+        d->rings[i]=NULL;
+      }
       // did the child to stop ?
       si_waitpid(d->pid,NULL,WNOHANG);
       if ((d->pid!=0)
@@ -1531,6 +1631,9 @@ leftv ssiRead1(si_link l)
     case 23: ssiReadRingProperties(l);
              return ssiRead1(l);
              break;
+    case 24: res->rtyp=BIGINTVEC_CMD;
+             res->data=ssiReadBigintvec(d);
+             break;
     // ------------
     case 98: // version
              {
@@ -1555,6 +1658,7 @@ leftv ssiRead1(si_link l)
                 return ssiRead1(l);
              }
     case 99: omFreeBin(res,sleftv_bin); ssiClose(l); m2_end(0);
+             break; /*to make compiler happy*/
     case 0: if (s_iseof(d->f_read))
             {
               ssiClose(l);
@@ -1580,7 +1684,7 @@ no_ring: WerrorS("no ring");
   return NULL;
 }
 //**************************************************************************/
-BOOLEAN ssiSetRing(si_link l, ring r, BOOLEAN send)
+static BOOLEAN ssiSetRing(si_link l, ring r, BOOLEAN send)
 {
   if(SI_LINK_W_OPEN_P(l)==0)
      if (slOpen(l,SI_LINK_OPEN|SI_LINK_WRITE,NULL)) return TRUE;
@@ -1725,6 +1829,10 @@ BOOLEAN ssiWrite(si_link l, leftv data)
                    fputs("19 ",d->f_write);
                    ssiWriteBigintmat(d,(bigintmat *)dd);
                    break;
+          case BIGINTVEC_CMD:
+                   fputs("24 ",d->f_write);
+                   ssiWriteBigintvec(d,(bigintmat *)dd);
+                   break;
           default:
             if (tt>MAX_TOK)
             {
@@ -1776,9 +1884,31 @@ const char* slStatusSsi(si_link l, const char* request)
   ||(strcmp(l->mode,"connect")==0))
   && (strcmp(request, "read") == 0))
   {
+    if (s_isready(d->f_read)) return "ready";
+#if defined(HAVE_POLL)
+    pollfd pfd;
+    loop
+    {
+      /* Don't block. Return socket status immediately. */
+      pfd.fd=d->fd_read;
+      pfd.events=POLLIN;
+      //Print("test fd %d\n",d->fd_read);
+    /* check with select: chars waiting: no -> not ready */
+      switch (si_poll(&pfd,1,0))
+      {
+        case 0: /* not ready */ return "not ready";
+        case -1: /*error*/      return "error";
+        case 1: /*ready ? */    break;
+      }
+#else
     fd_set  mask;
     struct timeval wt;
-    if (s_isready(d->f_read)) return "ready";
+    if (FD_SETSIZE<=d->fd_read)
+    {
+      Werror("file descriptor number too high (%d)",d->fd_read);
+      return "error";
+    }
+
     loop
     {
       /* Don't block. Return socket status immediately. */
@@ -1795,6 +1925,7 @@ const char* slStatusSsi(si_link l, const char* request)
         case -1: /*error*/      return "error";
         case 1: /*ready ? */    break;
       }
+#endif
     /* yes: read 1 char*/
     /* if \n, check again with select else ungetc(c), ready*/
       int c=s_getc(d->f_read);
@@ -1823,25 +1954,105 @@ const char* slStatusSsi(si_link l, const char* request)
   else return "unknown status request";
 }
 
-int slStatusSsiL(lists L, int timeout)
+int slStatusSsiL(lists L, int timeout, BOOLEAN *ignore)
 {
 // input: L: a list with links of type
 //           ssi-connect, ssi-fork, ssi-tcp, MPtcp-fork or MPtcp-launch.
 //           Note: Not every entry in L must be set.
-//        timeout: timeout for select in micro-seconds
+//        timeout: timeout for select in milli-seconds
 //           or -1 for infinity
 //           or 0 for polling
 // returns: ERROR (via Werror): L has wrong elements or link not open
-//           -2: select returns an error
+//           -2: error in L
 //           -1: the read state of all links is eof
 //           0:  timeout (or polling): none ready,
 //           i>0: (at least) L[i] is ready
+#if defined(HAVE_POLL) && !defined(__APPLE__)
   si_link l;
-  ssiInfo *d;
+  ssiInfo *d=NULL;
   int d_fd;
-  fd_set  mask, fdmask;
+  int s;
+  int nfd=L->nr+1;
+  pollfd *pfd=(pollfd*)omAlloc0(nfd*sizeof(pollfd));
+  for(int i=L->nr; i>=0; i--)
+  {
+    pfd[i].fd=-1;
+    if (L->m[i].Typ()!=DEF_CMD)
+    {
+      if (L->m[i].Typ()!=LINK_CMD)
+      { WerrorS("all elements must be of type link"); return -2;}
+      l=(si_link)L->m[i].Data();
+      if(SI_LINK_OPEN_P(l)==0)
+      { WerrorS("all links must be open"); return -2;}
+      if (((strcmp(l->m->type,"ssi")!=0) && (strcmp(l->m->type,"MPtcp")!=0))
+      || ((strcmp(l->mode,"fork")!=0) && (strcmp(l->mode,"tcp")!=0)
+        && (strcmp(l->mode,"launch")!=0) && (strcmp(l->mode,"connect")!=0)))
+      {
+        WerrorS("all links must be of type ssi:fork, ssi:tcp, ssi:connect");
+        return -2;
+      }
+      if (strcmp(l->m->type,"ssi")==0)
+      {
+        d=(ssiInfo*)l->data;
+        d_fd=d->fd_read;
+        if (!s_isready(d->f_read))
+        {
+          pfd[i].fd=d_fd;
+          pfd[i].events=POLLIN;
+        }
+        else
+        {
+          return i+1;
+        }
+      }
+      else
+      {
+        Werror("wrong link type >>%s<<",l->m->type);
+        return -2;
+      }
+    }
+    else if (ignore!=NULL)
+    {
+      ignore[i]=TRUE; // not a link
+    }
+  }
+  s=si_poll(pfd,nfd,timeout);
+  if (s==-1)
+  {
+    Werror("error in poll call (errno:%d)",errno);
+    return -2; /*error*/
+  }
+  if(s==0)
+  {
+    return 0; /*timeout*/
+  }
+  for(int i=L->nr; i>=0; i--)
+  {
+    if ((L->m[i].rtyp==LINK_CMD)
+    && ((ignore==NULL)||(ignore[i]==FALSE)))
+    {
+      // the link type is ssi, that's already tested
+      l=(si_link)L->m[i].Data();
+      d=(ssiInfo*)l->data;
+      d_fd=d->fd_read;
+      if (pfd[i].fd==d_fd)
+      {
+        if (pfd[i].revents &POLLIN)
+        {
+          omFree(pfd);
+          return i+1;
+        }
+      }
+    }
+  }
+  // no ready
+  return 0;
+#else
+  si_link l;
+  ssiInfo *d=NULL;
+  int d_fd;
+  fd_set fdmask;
   FD_ZERO(&fdmask);
-  FD_ZERO(&mask);
   int max_fd=0; /* 1 + max fd in fd_set */
 
   /* timeout */
@@ -1854,8 +2065,8 @@ int slStatusSsiL(lists L, int timeout)
   }
   else
   {
-    wt.tv_sec  = timeout / 1000000;
-    wt.tv_usec = timeout % 1000000;
+    wt.tv_sec  = timeout / 1000;
+    wt.tv_usec = (timeout % 1000)*1000;
   }
 
   /* auxiliary variables */
@@ -1889,8 +2100,11 @@ int slStatusSsiL(lists L, int timeout)
         d_fd=d->fd_read;
         if (!s_isready(d->f_read))
         {
-          FD_SET(d_fd, &fdmask);
-          if (d_fd > max_fd) max_fd=d_fd;
+          if ((ignore==NULL) || (ignore[i]==FALSE))
+          {
+            FD_SET(d_fd, &fdmask);
+            if (d_fd > max_fd) max_fd=d_fd;
+          }
         }
         else
           return i+1;
@@ -1903,23 +2117,17 @@ int slStatusSsiL(lists L, int timeout)
     }
   }
   max_fd++;
-
-do_select:
-  /* copy fdmask to mask */
-  FD_ZERO(&mask);
-  for(k = 0; k < max_fd; k++)
+  if (FD_SETSIZE<=max_fd)
   {
-    if(FD_ISSET(k, &fdmask))
-    {
-      FD_SET(k, &mask);
-    }
+    Werror("file descriptor number too high (%d)",max_fd);
+    return -2;
   }
 
   /* check with select: chars waiting: no -> not ready */
-  s = si_select(max_fd, &mask, NULL, NULL, wt_ptr);
+  s = si_select(max_fd, &fdmask, NULL, NULL, wt_ptr);
   if (s==-1)
   {
-    WerrorS("error in select call");
+    Werror("error in select call (errno:%d)",errno);
     return -2; /*error*/
   }
   if (s==0)
@@ -1929,7 +2137,7 @@ do_select:
   else /* s>0, at least one ready  (the number of fd which are ready is s)*/
   {
     j=0;
-    while (j<=max_fd) { if (FD_ISSET(j,&mask)) break; j++; }
+    while (j<=max_fd) { if (FD_ISSET(j,&fdmask)) break; j++; }
     for(i=L->nr; i>=0; i--)
     {
       if (L->m[i].rtyp==LINK_CMD)
@@ -1939,60 +2147,13 @@ do_select:
         {
           d=(ssiInfo*)l->data;
           d_fd=d->fd_read;
-          if(j==d_fd) break;
-        }
-        else
-        {
-          Werror("wrong link type >>%s<<",l->m->type);
-          return -2;
+          if(j==d_fd) return i+1;
         }
       }
     }
-    // only ssi links:
-    loop
-    {
-      /* yes: read 1 char*/
-      /* if \n, check again with select else ungetc(c), ready*/
-      /* setting: d: current ssiInfo, j current fd, i current entry in L*/
-      int c=s_getc(d->f_read);
-      //Print("try c=%d\n",c);
-      if (c== -1) /* eof */
-      {
-        FD_CLR(j,&fdmask);
-        fdmaskempty = 1;
-        for(k = 0; k < max_fd; k++)
-        {
-          if(FD_ISSET(k, &fdmask))
-          {
-            fdmaskempty = 0;
-            break;
-          }
-        }
-        if(fdmaskempty)
-        {
-          return -1;
-        }
-        if(timeout != -1)
-        {
-          timeout = si_max(0,
-             timeout - 1000000*(getRTimer()/TIMER_RESOLUTION - startingtime));
-          wt.tv_sec  = timeout / 1000000;
-          wt.tv_usec = (timeout % 1000000);
-        }
-        goto do_select;
-      }
-
-      else if (isdigit(c))
-      { s_ungetc(c,d->f_read); return i+1; }
-      else if (c>' ')
-      {
-        Werror("unknown char in ssiLink(%d)",c);
-        return -2;
-      }
-      /* else: next char */
-      goto do_select;
-    }
   }
+  return 0;
+#endif
 }
 
 int ssiBatch(const char *host, const char * port)
@@ -2000,7 +2161,7 @@ int ssiBatch(const char *host, const char * port)
 {
   si_link l=(si_link) omAlloc0Bin(sip_link_bin);
   char *buf=(char*)omAlloc(256);
-  sprintf(buf,"ssi:connect %s:%s",host,port);
+  snprintf(buf,256,"ssi:connect %s:%s",host,port);
   slInit(l, buf);
   omFreeSize(buf,256);
   if (slOpen(l,SI_LINK_OPEN,NULL)) return 1;
@@ -2341,6 +2502,7 @@ BOOLEAN ssiGetDump(si_link l)
 // 23 0 <log(bitmask)> ring properties: max.exp.
 // 23 1 <log(bitmask)> <r->IsLPRing> ring properties:LPRing
 // 23 2 <matrix C> <matrix D> ring properties: PLuralRing
+// 24 bigintvec <c>
 //
 // 98: verify version: <ssi-version> <MAX_TOK> <OPT1> <OPT2>
 // 99: quit Singular
