@@ -38,7 +38,7 @@
 #include "feOpt.h"
 #include "links/silink.h"
 #include "mod_lib.h"
-#include "Singular/distrib.h"
+#include "misc/distrib.h"
 
 #include "misc/options.h"
 #include "misc/intvec.h"
@@ -68,15 +68,7 @@
 
 #ifdef HAVE_NTL
 #include <NTL/version.h>
-#include <NTL/tools.h>
-#ifdef NTL_CLIENT
-NTL_CLIENT
 #endif
-#endif
-
-
-static FORCE_INLINE void number2mpz(number n, mpz_t m){ number2mpz(n, coeffs_BIGINT, m); }
-static FORCE_INLINE number mpz2number(mpz_t m){ return mpz2number(m, coeffs_BIGINT); }
 
 
 void setListEntry(lists L, int index, mpz_t n)
@@ -92,7 +84,7 @@ void setListEntry(lists L, int index, mpz_t n)
       return;
     }
   }
-  number nn = mpz2number(n);
+  number nn = mpz2number(n, coeffs_BIGINT);
   L->m[index].rtyp = BIGINT_CMD; L->m[index].data = (void*)nn;
 }
 
@@ -366,7 +358,7 @@ lists primeFactorisation(const number n, const int pBound)
 {
   int i;
   int index=0;
-  mpz_t nn; number2mpz(n, nn);
+  mpz_t nn; number2mpz(n, coeffs_BIGINT, nn);
   lists primes = (lists)omAllocBin(slists_bin); primes->Init(1000);
   int* multiplicities = (int*)omAlloc0(1000*sizeof(int));
   int positive=1;
@@ -474,7 +466,7 @@ void singular_example(char *str)
     char *res_m=feResource('m', 0);
     if (res_m!=NULL)
     {
-      sprintf(sing_file, "%s/%s.sing", res_m, s);
+      snprintf(sing_file,MAXPATHLEN, "%s/%s.sing", res_m, s);
       fd = feFopen(sing_file, "r");
     }
     if (fd != NULL)
@@ -703,7 +695,7 @@ BOOLEAN setOption(leftv res, leftv v)
   okay:
     if (currRing != NULL)
       currRing->options = si_opt_1 & TEST_RINGDEP_OPTS;
-    omFree((ADDRESS)n);
+    omFreeBinAddr((ADDRESS)n);
     v=v->next;
   } while (v!=NULL);
 
@@ -772,7 +764,7 @@ extern "C"
 #endif
 
 #ifndef MAKE_DISTRIBUTION
-const char *singular_date=__DATE__ " " __TIME__;
+const char *singular_date = __DATE__ " " __TIME__;
 #endif
 
 char * versionString(/*const bool bShowDetails = false*/ )
@@ -837,7 +829,11 @@ char * versionString(/*const bool bShowDetails = false*/ )
               StringAppendS("Plural,");
 #endif
 #ifdef HAVE_VSPACE
-              StringAppendS("vspace,");
+  #if defined(__GNUC__) && (__GNUC__<9) &&!defined(__clang__)
+              StringAppendS("vspace(1),");
+  #else
+              StringAppendS("vspace(2),");
+  #endif
 #endif
 #ifdef HAVE_DBM
               StringAppendS("DBM,\n\t");
@@ -873,6 +869,9 @@ char * versionString(/*const bool bShowDetails = false*/ )
 #endif
 #ifdef KDEBUG
               StringAppendS("KDEBUG,");
+#endif
+#ifdef HAVE_SDB
+              StringAppendS("sdb,");
 #endif
               StringAppendS("\n\t");
 #ifdef __OPTIMIZE__
@@ -884,6 +883,11 @@ char * versionString(/*const bool bShowDetails = false*/ )
 #ifdef __NO_INLINE__
               StringAppendS("CC:NO_INLINE,");
 #endif
+#ifdef HAVE_NTL
+  #ifdef NTL_AVOID_BRANCHING
+  #undef HAVE_GENERIC_ADD
+  #endif
+#endif
 #ifdef HAVE_GENERIC_ADD
               StringAppendS("GenericAdd,");
 #else
@@ -1098,7 +1102,21 @@ void m2_end(int i)
   if (!m2_end_called)
   {
     EXTERN_VAR FILE* File_Profiling;
+    EXTERN_VAR FILE* File_Log;
+    EXTERN_VAR BOOLEAN File_Log_written;
     if (File_Profiling!=NULL) { fclose(File_Profiling); File_Profiling=NULL; }
+    if (File_Log!=NULL)
+    {
+      fclose(File_Log);
+      File_Log=NULL;
+      if (File_Log_written==FALSE) // remove empty logs
+      {
+        int pid=getpid();
+        char buf[20];
+        snprintf(buf,20,"/tmp/sing_log.%d",pid);
+        remove(buf);
+      }
+    }
     m2_end_called = TRUE;
 #ifdef HAVE_SIMPLEIPC
     for (int j = SIPC_MAX_SEMAPHORES-1; j >= 0; j--)
@@ -1117,7 +1135,6 @@ void m2_end(int i)
       }
     }
 #endif   // HAVE_SIMPLEIPC
-    fe_reset_input_mode();
     monitor(NULL,0);
 #ifdef PAGE_TEST
     mmEndStat();
