@@ -15,6 +15,7 @@
 
 #include <config.h>
 
+#include <string.h>
 
 #include "canonicalform.h"
 #include "fac_util.h"
@@ -62,6 +63,7 @@ extern "C"
 #endif
 #if ( __FLINT_RELEASE >= 20503)
 #include <flint/fmpq_mpoly.h>
+#include <flint/fmpz_mod.h>
 
 // planed, but not yet in FLINT:
 #if (__FLINT_RELEASE < 20700)
@@ -85,7 +87,7 @@ void fq_nmod_set_nmod_poly(fq_nmod_t a, const nmod_poly_t b, const fq_nmod_ctx_t
     nmod_poly_set(a, b);
     fq_nmod_reduce(a, ctx);
 }
-#else
+#elif  (__FLINT_RELEASE < 30000)
 void fq_nmod_set_nmod_poly(fq_nmod_t a, const nmod_poly_t b,
                                                        const fq_nmod_ctx_t ctx)
 {
@@ -478,7 +480,7 @@ convertFacCF2Fq_nmod_t (fq_nmod_t result, const CanonicalForm& f,
 }
 
 CanonicalForm
-convertFq_nmod_t2FacCF (const fq_nmod_t poly, const Variable& alpha, const fq_nmod_ctx_t ctx)
+convertFq_nmod_t2FacCF (const fq_nmod_t poly, const Variable& alpha, const fq_nmod_ctx_t /*ctx*/)
 {
   return convertnmod_poly_t2FacCF (poly, alpha);
 }
@@ -1006,7 +1008,7 @@ CanonicalForm gcdFlintMP_QQ(const CanonicalForm& F, const CanonicalForm& G)
 #if __FLINT_RELEASE >= 20700
 CFFList
 convertFLINTFq_nmod_mpoly_factor2FacCFFList (
-                   fq_nmod_mpoly_factor_t fac, 
+                   fq_nmod_mpoly_factor_t fac,
                    const fq_nmod_mpoly_ctx_t& ctx,
                    const int N,
                    const fq_nmod_ctx_t& fq_ctx,
