@@ -21,45 +21,36 @@
 
 
 
-#include <misc/auxiliary.h>
+#include "misc/auxiliary.h"
 
 #ifdef HAVE_PLURAL
 
 // for
 #define PLURAL_INTERNAL_DECLARATIONS
 
-#include <polys/nc/sca.h>
-#include <polys/nc/nc.h>
-#include <polys/nc/gb_hack.h>
-// #include <polys/gring.h>
+#include "polys/nc/sca.h"
+#include "polys/nc/nc.h"
+#include "polys/nc/gb_hack.h"
 
+#include "coeffs/numbers.h"
 
-#include <coeffs/numbers.h>
+#include "misc/options.h"
 
-// #include <polys/febase.h>
-#include <misc/options.h>
+#include "polys/monomials/p_polys.h"
 
-#include <polys/monomials/p_polys.h>
+#include "polys/simpleideals.h"
+#include "misc/intvec.h"
 
-// #include <polys/kutil.h>
-#include <polys/simpleideals.h>
-#include <misc/intvec.h>
+#include "polys/monomials/ring.h"
+#include "polys/kbuckets.h"
+#include "polys/sbuckets.h"
 
-#include <polys/monomials/ring.h>
-#include <polys/kbuckets.h>
-// #include <polys/kstd1.h>
-#include <polys/sbuckets.h>
+#include "polys/prCopy.h"
 
-#include <polys/prCopy.h>
-
-#include <polys/operations/p_Mult_q.h>
-#include <polys/templates/p_MemAdd.h>
-
-// #include <polys/kutil.h>
-// #include <polys/kstd1.h>
-
-#include <polys/weight.h>
+#include "polys/operations/p_Mult_q.h"
+#include "polys/templates/p_MemAdd.h"
 
+#include "polys/weight.h"
 
 // poly functions defined in p_Procs :
 
@@ -67,13 +58,13 @@
 poly sca_pp_Mult_mm(const poly pPoly, const poly pMonom, const ring rRing, poly &);
 
 // return pMonom * pPoly; preserve pPoly and pMonom.
-static poly sca_mm_Mult_pp(const poly pMonom, const poly pPoly, const ring rRing);
+static poly sca_pp_mm_Mult(const poly pPoly, const poly pMonom, const ring rRing);
 
 // return pPoly * pMonom; preserve pMonom, destroy or reuse pPoly.
 poly sca_p_Mult_mm(poly pPoly, const poly pMonom, const ring rRing);
 
 // return pMonom * pPoly; preserve pMonom, destroy or reuse pPoly.
-static poly sca_mm_Mult_p(const poly pMonom, poly pPoly, const ring rRing);
+static poly sca_p_mm_Mult(poly pPoly, const poly pMonom, const ring rRing);
 
 
 // compute the spolynomial of p1 and p2
@@ -97,10 +88,10 @@ static inline int sca_Sign_mm_Mult_mm( const poly pMonomM, const poly pMonomMM,
     const short iFirstAltVar = scaFirstAltVar(rRing);
     const short iLastAltVar  = scaLastAltVar(rRing);
 
-    register unsigned int tpower = 0;
-    register unsigned int cpower = 0;
+    REGISTER unsigned int tpower = 0;
+    REGISTER unsigned int cpower = 0;
 
-    for( register short j = iLastAltVar; j >= iFirstAltVar; j-- )
+    for( REGISTER short j = iLastAltVar; j >= iFirstAltVar; j-- )
     {
       const unsigned int iExpM  = p_GetExp(pMonomM,  j, rRing);
       const unsigned int iExpMM = p_GetExp(pMonomMM, j, rRing);
@@ -148,10 +139,10 @@ static inline poly sca_m_Mult_mm( poly pMonomM, const poly pMonomMM, const ring
     const unsigned int iFirstAltVar = scaFirstAltVar(rRing);
     const unsigned int iLastAltVar = scaLastAltVar(rRing);
 
-    register unsigned int tpower = 0;
-    register unsigned int cpower = 0;
+    REGISTER unsigned int tpower = 0;
+    REGISTER unsigned int cpower = 0;
 
-    for( register unsigned int j = iLastAltVar; j >= iFirstAltVar; j-- )
+    for( REGISTER unsigned int j = iLastAltVar; j >= iFirstAltVar; j-- )
     {
       const unsigned int iExpM  = p_GetExp(pMonomM,  j, rRing);
       const unsigned int iExpMM = p_GetExp(pMonomMM, j, rRing);
@@ -212,10 +203,10 @@ static inline poly sca_mm_Mult_m( const poly pMonomMM, poly pMonomM, const ring
     const unsigned int iFirstAltVar = scaFirstAltVar(rRing);
     const unsigned int iLastAltVar = scaLastAltVar(rRing);
 
-    register unsigned int tpower = 0;
-    register unsigned int cpower = 0;
+    REGISTER unsigned int tpower = 0;
+    REGISTER unsigned int cpower = 0;
 
-    for( register unsigned int j = iLastAltVar; j >= iFirstAltVar; j-- )
+    for( REGISTER unsigned int j = iLastAltVar; j >= iFirstAltVar; j-- )
     {
       const unsigned int iExpMM = p_GetExp(pMonomMM, j, rRing);
       const unsigned int iExpM  = p_GetExp(pMonomM,  j, rRing);
@@ -277,10 +268,10 @@ static inline poly sca_mm_Mult_mm( poly pMonom1, const poly pMonom2, const ring
     const unsigned int iFirstAltVar = scaFirstAltVar(rRing);
     const unsigned int iLastAltVar = scaLastAltVar(rRing);
 
-    register unsigned int tpower = 0;
-    register unsigned int cpower = 0;
+    REGISTER unsigned int tpower = 0;
+    REGISTER unsigned int cpower = 0;
 
-    for( register unsigned int j = iLastAltVar; j >= iFirstAltVar; j-- )
+    for( REGISTER unsigned int j = iLastAltVar; j >= iFirstAltVar; j-- )
     {
       const unsigned int iExp1 = p_GetExp(pMonom1, j, rRing);
       const unsigned int iExp2 = p_GetExp(pMonom2, j, rRing);
@@ -346,9 +337,9 @@ static inline poly sca_xi_Mult_mm(short i, const poly pMonom, const ring rRing)
 
     const short iFirstAltVar = scaFirstAltVar(rRing);
 
-    register unsigned int cpower = 0;
+    REGISTER unsigned int cpower = 0;
 
-    for( register short j = iFirstAltVar; j < i ; j++ )
+    for( REGISTER short j = iFirstAltVar; j < i ; j++ )
       cpower ^= p_GetExp(pMonom, j, rRing);
 
 #ifdef PDEBUG
@@ -573,7 +564,7 @@ static inline poly sca_xi_Mult_pp(short i, const poly pPoly, const ring rRing)
 
 
 // return new poly = pMonom * pPoly; preserve pPoly and pMonom.
-static poly sca_mm_Mult_pp(const poly pMonom, const poly pPoly, const ring rRing)
+static poly sca_pp_mm_Mult(const poly pPoly, const poly pMonom, const ring rRing)
 {
   assume( rIsSCA(rRing) );
 
@@ -605,7 +596,7 @@ static poly sca_mm_Mult_pp(const poly pMonom, const poly pPoly, const ring rRing
       if( iComponent!=0 ) // TODO: make global if on iComponentMonomM =?= 0
       {
         // REPORT_ERROR
-        Werror("sca_mm_Mult_pp: exponent mismatch %d and %d\n", iComponent, iComponentMonomM);
+        Werror("sca_pp_mm_Mult: exponent mismatch %d and %d\n", iComponent, iComponentMonomM);
         // what should we do further?!?
 
         p_Delete( &pResult, rRing); // delete the result
@@ -614,7 +605,7 @@ static poly sca_mm_Mult_pp(const poly pMonom, const poly pPoly, const ring rRing
 #ifdef PDEBUG
       if(iComponent==0 )
       {
-        dReportError("sca_mm_Mult_pp: Multiplication in the left module from the right!");
+        dReportError("sca_pp_mm_Mult: Multiplication in the left module from the right!");
 //        PrintS("mm = "); p_Write(pMonom, rRing);
 //        PrintS("pp = "); p_Write(pPoly, rRing);
 //        assume(iComponent!=0);
@@ -641,7 +632,7 @@ static poly sca_mm_Mult_pp(const poly pMonom, const poly pPoly, const ring rRing
 
 
 // return poly = pMonom * pPoly; preserve pMonom, destroy or reuse pPoly.
-static poly sca_mm_Mult_p(const poly pMonom, poly pPoly, const ring rRing) // !!!!! the MOST used procedure !!!!!
+static poly sca_p_mm_Mult(poly pPoly, const poly pMonom, const ring rRing) // !!!!! the MOST used procedure !!!!!
 {
   assume( rIsSCA(rRing) );
 
@@ -682,7 +673,7 @@ static poly sca_mm_Mult_p(const poly pMonom, poly pPoly, const ring rRing) // !!
       if( iComponent!=0 )
       {
         // REPORT_ERROR
-        Werror("sca_mm_Mult_p: exponent mismatch %d and %d\n", iComponent, iComponentMonomM);
+        Werror("sca_p_mm_Mult: exponent mismatch %d and %d\n", iComponent, iComponentMonomM);
         // what should we do further?!?
 
         p_Delete( &pPoly, rRing); // delete the result
@@ -691,7 +682,7 @@ static poly sca_mm_Mult_p(const poly pMonom, poly pPoly, const ring rRing) // !!
 #ifdef PDEBUG
       if(iComponent==0)
       {
-        dReportError("sca_mm_Mult_p: Multiplication in the left module from the right!");
+        dReportError("sca_p_mm_Mult: Multiplication in the left module from the right!");
 //        PrintS("mm = "); p_Write(pMonom, rRing);
 //        PrintS("p = "); p_Write(pPoly, rRing);
 //        assume(iComponent!=0);
@@ -765,7 +756,7 @@ poly sca_SPoly( const poly p1, const poly p2, const ring r )
     return(NULL);
   }
 
-  poly pL = p_Lcm(p1, p2, si_max(lCompP1, lCompP2), r);       // pL = lcm( lm(p1), lm(p2) )
+  poly pL = p_Lcm(p1, p2, r);       // pL = lcm( lm(p1), lm(p2) )
 
   poly m1 = p_One( r);
   p_ExpVectorDiff(m1, pL, p1, r);                  // m1 = pL / lm(p1)
@@ -821,7 +812,6 @@ poly sca_SPoly( const poly p1, const poly p2, const ring r )
   poly spoly = p_Add_q (tmp1, tmp2, r); // spoly = spoly(lt(p1), lt(p2)) + m1 * tail(p1), delete tmp1,2
 
   if (spoly!=NULL) p_Cleardenom (spoly, r);
-//  if (spoly!=NULL) p_Content (spoly); // r?
 
 #ifdef PDEBUG
   p_Test (spoly, r);
@@ -896,7 +886,7 @@ poly sca_ReduceSpoly(const poly p1, poly p2, const ring r)
 
   p2 = p_Add_q(p2, T, r);
 
-  if ( p2!=NULL ) p_Content(p2,r);
+  if ( p2!=NULL ) p_Cleardenom(p2,r);
 
 #ifdef PDEBUG
   p_Test(p2,r);
@@ -954,10 +944,6 @@ bool sca_SetupQuotient(ring rGR, ring rG, bool bCopy)
 
   const int N = rG->N;
 
-  if(N < 2)
-    return false;
-
-
 //  if( (ncRingType(rG) != nc_skew) || (ncRingType(rG) != nc_comm) )
 //    return false;
 
@@ -1029,9 +1015,11 @@ bool sca_SetupQuotient(ring rGR, ring rG, bool bCopy)
 #endif
 
 
-  if( (iAltVarEnd == -1) || (iAltVarStart == (N+1)) )
-    return false; // either no alternating varables, or a single one => we are in commutative case!
-
+  if((iAltVarEnd == -1) || (iAltVarStart == (N+1)))
+  {
+    if (N>1) return false; // no alternating varables
+    iAltVarEnd=iAltVarStart=1;
+  }
 
   for(int i = 1; i < N; i++)
   {
@@ -1068,7 +1056,7 @@ bool sca_SetupQuotient(ring rGR, ring rG, bool bCopy)
 #endif
 
   assume( 1            <= iAltVarStart );
-  assume( iAltVarStart < iAltVarEnd   );
+  assume( iAltVarStart <= iAltVarEnd   );
   assume( iAltVarEnd   <= N            );
 
 
@@ -1092,8 +1080,6 @@ bool sca_SetupQuotient(ring rGR, ring rG, bool bCopy)
   // y_{iAltVarStart}^2, y_{iAltVarStart+1}^2, \ldots, y_{iAltVarEnd}^2  (iAltVarEnd > iAltVarStart)
   // to be within quotient ideal.
 
-  bool bSCA = true;
-
   int b = N+1;
   int e = -1;
 
@@ -1107,7 +1093,7 @@ bool sca_SetupQuotient(ring rGR, ring rG, bool bCopy)
 #endif
   }
 
-  for ( int i = iAltVarStart; (i <= iAltVarEnd) && bSCA; i++ )
+  for ( int i = iAltVarStart; (i <= iAltVarEnd); i++ )
     if( (i < b) || (i > e) ) // otherwise it's ok since rG is an SCA!
     {
       poly square = p_One( rG);
@@ -1121,8 +1107,7 @@ bool sca_SetupQuotient(ring rGR, ring rG, bool bCopy)
       if( square != NULL ) // var(i)^2 is not in Q?
       {
         p_Delete(&square, rG);
-        bSCA = false;
-        break;
+        return false;
       }
     }
 
@@ -1246,8 +1231,8 @@ void sca_p_ProcsSet(ring rGR, p_Procs_s* p_Procs)
   p_Procs->pp_Mult_mm         = sca_pp_Mult_mm;
 
   // non-commutaitve
-  rGR->GetNC()->p_Procs.mm_Mult_p   = sca_mm_Mult_p;
-  rGR->GetNC()->p_Procs.mm_Mult_pp  = sca_mm_Mult_pp;
+  p_Procs->p_mm_Mult          = sca_p_mm_Mult;
+  p_Procs->pp_mm_Mult         = sca_pp_mm_Mult;
 
 //   rGR->GetNC()->p_Procs.SPoly         = sca_SPoly;
 //   rGR->GetNC()->p_Procs.ReduceSPoly   = sca_ReduceSpoly;
